# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from PyQt5.Qt import QObject, Qt, QActionGroup
from PyQt5.QtCore import pyqtSignal

import TigGUI.kitties.utils
from TigGUI.init import Config, ConfigFile, ConfigFileName, pixmaps
from TigGUI.kitties.utils import PersistentCurrier

_verbosity = TigGUI.kitties.utils.verbosity(name="mmod")
dprint = _verbosity.dprint
dprintf = _verbosity.dprintf

_Contexts = dict(image=1, model=2)

WHEELUP = "WheelUp"
WHEELDOWN = "WheelDown"

MM_ZWIN = "zoom-window"
MM_ZUNDO = "zoom-undo"
MM_ZREDO = "zoom-redo"
MM_UNZOOM = "unzoom"
MM_MEAS = "measure"
MM_STATS = "stats"
MM_SELSRC = "select-source"
MM_SELWIN = "select-window"
MM_SELWINPLUS = "select-window-plus"
MM_DESEL = "deselect-window"

_AllFuncs = [MM_ZWIN, MM_ZUNDO, MM_ZREDO, MM_UNZOOM,
             MM_MEAS, MM_STATS, MM_SELSRC, MM_SELWIN, MM_SELWINPLUS, MM_DESEL]

FuncDoc = {
    MM_ZWIN: "Zoom into window (click-drag) or zoom in at point (double-click)",
    MM_ZUNDO: "Zoom out to previous view",
    MM_ZREDO: "Zoom in to previous view",
    MM_UNZOOM: "Zoom fully out",
    MM_MEAS: "Measuring ruler (click and drag)",
    MM_STATS: "Stats in box (click and drag)",
    MM_SELSRC: "Select nearest source",
    MM_SELWIN: "Select sources in window",
    MM_SELWINPLUS: "Extend selection with sources in window",
    MM_DESEL: "Deselect sources in window"
}

_DefaultModes = "Mouse3,Mouse2,Mouse1"
_DefaultInitialMode = "Mouse3"


class MouseModeManager(QObject):
    class MouseMode:

        def __init__(self, mid):
            self.id = mid
            self.name = self.icon = self.tooltip = None
            self.contexts = []
            self.submodes = []
            self.patterns = {}
            self.qa = None

        def addAction(self, menu, qag, callback, toolbar=None):
            self.qa = menu.addAction(self.name, callback)
            icon = self.icon and getattr(pixmaps, self.icon, None)
            icon and self.qa.setIcon(icon.icon())
            self.qa.setCheckable(True)
            qag.addAction(self.qa)
            toolbar and toolbar.addAction(self.qa)

    setMouseMode = pyqtSignal(MouseMode)

    def __init__(self, parent, menu, toolbar):
        QObject.__init__(self, parent)
        self._currier = PersistentCurrier()
        # get list of mouse modes from config
        modelist = []
        for mid in Config.get("mouse-modes", _DefaultModes).split(","):
            if not ConfigFile.has_section(mid):
                print("ERROR: unknown mouse-mode '%s', skipping. Check your %s." % (mid, ConfigFileName))
            else:
                modelist.append(self._readModeConfig(mid))
        self._modes = dict([(mode.id, mode) for mode in modelist])
        self._qag_mode = QActionGroup(self)
        self._qag_submode = QActionGroup(self)
        self._all_submodes = []
        # make entries for main modes
        for mode in modelist:
            mode.addAction(menu, self._qag_mode, callback=self._currier.curry(self._setMode, mode.id))
            if mode.submodes:
                self._all_submodes += list(mode.submodes)
        # make entries for submodes
        self._qa_submode_sep = menu.addSeparator()
        self._modes.update([(mode.id, mode) for mode in self._all_submodes])
        for mode in self._all_submodes:
            mode.addAction(menu, self._qag_submode, toolbar=toolbar,
                           callback=self._currier.curry(self._setSubmode, mode.id))
        # other init
        self._current_context = None
        self._available_submodes = []
        # set initial mode
        initmode = Config.get("current-mouse-mode", _DefaultInitialMode)
        if initmode not in self._modes:
            initmode = modelist[0].id
        self._modes[initmode].qa.setChecked(True)
        self._setMode(initmode, write_config=False)

    def currentMode(self):
        return self._current_submode or self._current_mode

    def setContext(self, has_image, has_model):
        self._current_context = (has_image and _Contexts['image']) | (has_model and _Contexts['model'])
        self._ensureValidSubmodes()

    def _ensureValidSubmodes(self):
        current = None
        self._valid_submodes = []
        # accumulate list of valid submodes, and find the checked-on one
        for mode in self._available_submodes:
            if not mode.contexts or not self._current_context or self._current_context & mode.contexts:
                self._valid_submodes.append(mode)
                mode.qa.setVisible(True)
                if mode.qa.isChecked():
                    current = mode.id
            else:
                mode.qa.setVisible(False)
        if self._valid_submodes:
            self._setSubmode(current or self._valid_submodes[0].id)

    def _setMode(self, mid, write_config=True):
        """Called when the mouse mode changes"""
        if write_config:
            Config.set("current-mouse-mode", mid)
        self._current_mode = mode = self._modes[mid]
        # hide submodes if any
        for mm in self._all_submodes:
            mm.qa.setVisible(False)
        self._qa_submode_sep.setVisible(bool(mode.submodes))
        self._current_submode = None
        self._available_submodes = mode.submodes
        # make relevant submodes visible, and make sure one is enabled
        if mode.submodes:
            self._ensureValidSubmodes()
        else:
            self.setMouseMode.emit(mode)

    def _setSubmode(self, mid):
        """Called when the mouse submode changes"""
        self._current_submode = mode = self._modes[mid]
        mode.qa.setChecked(True)
        # hide submodes if any
        for mm in self._all_submodes:
            mm.qa.setShortcuts([])
        # set F4 shortcut to next submode
        if len(self._valid_submodes) > 1:
            for i, mm in enumerate(self._valid_submodes):
                if mm is mode:
                    self._valid_submodes[(i + 1) % len(self._valid_submodes)].qa.setShortcut(Qt.Key_F4)
                    break
        self.setMouseMode.emit(mode)

    def _readModeConfig(self, section, main_tooltip=None):
        """Reads the given config section (and uses the supplied defaults dict)
        and returns a dict of mouse_patterns,key_patterns per each function."""
        # read basic stuff
        mode = self.MouseMode(section)
        config = TigGUI.kitties.config.SectionParser(ConfigFile, section)
        mode.name = config.get("name", section)
        mode.icon = config.get("icon", "") or None
        mode.contexts = sum([_Contexts.get(x, 0) for x in config.get("contexts", "").split(",")])
        submodes = config.get("submodes", "") or None
        # eiher a mode with submodes, or a main mode
        if submodes:
            mode.tooltip = "<P>Your current mouse scheme is \"%s\".</P>" % mode.name
            for mid in submodes.split(","):
                if ConfigFile.has_section(mid):
                    mode.submodes.append(self._readModeConfig(mid, main_tooltip=mode.tooltip))
                else:
                    print("ERROR: unknown submode '%s' in mode config section '%s', skipping/ Check your %s." % (
                        mid, section, ConfigFileName))
        else:
            if main_tooltip:
                mode.tooltip = main_tooltip + """<P>In this scheme, available mouse functions depend on the selected mode.
        The current mode is %s. Use F4 to cycle through other modes.</P>""" % mode.name
            else:
                mode.tooltip = "<P>Your current mouse scheme is: \"%s\".</P>" % mode.name
            mode.tooltip += """<P>The following mouse functions are available:</P><BR><TABLE>\n"""
            patterns = {}
            # get basic patterns
            for func in _AllFuncs:
                # get pattern
                pattern = config.get(func, "")
                if not pattern:
                    continue
                mouse_pattern = key_pattern = None
                for pat in pattern.split(";"):
                    pat = pat.strip()
                    if pat and pat.lower() != "none":
                        # split by "+" and lookup each identifier in the Qt namespace
                        scomps = pat.split("+")
                        try:
                            comps = [x if x in (WHEELUP, WHEELDOWN) else getattr(Qt, x) for x in scomps]
                        except AttributeError:
                            print(
                                "WARNING: can't parse '%s' for function '%s' in mode config section '%s', disabling. Check your %s." % (
                                    pat, func, section, ConfigFileName))
                            continue
                        # append key/button code and sum of modifiers to the key or keyboard pattern list
                        if scomps[-1].startswith("Key_"):
                            if key_pattern:
                                print(
                                    "WARNING: more than one key pattern for function '%s' in mode config section '%s', ignoring. Check your %s." % (
                                        func, section, ConfigFileName))
                            else:
                                key_pattern = comps[-1], sum(comps[:-1])
                        else:
                            if mouse_pattern:
                                print(
                                    "WARNING: more than one mouse pattern for function '%s' in mode config section '%s', ignoring. Check your %s." % (
                                        func, section, ConfigFileName))
                            else:
                                mouse_pattern = comps[-1], sum(comps[:-1])
                mode.tooltip += "<TR><TD>%s:&nbsp;&nbsp;</TD><TD>%s</TD></TR>\n" % (pattern, FuncDoc[func])
                mode.patterns[func] = (mouse_pattern or (0, 0), key_pattern or (0, 0))
            mode.tooltip += "</TABLE><BR>"
        return mode



#!/usr/bin/env python3

# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import sys
from configparser import ConfigParser, NoSectionError, NoOptionError, DuplicateSectionError

import os.path

_default_system_paths = [
    "/usr/local/Timba",
    "/usr/Timba/",
    "/etc/"]

_default_user_path = os.path.expanduser("~/")


class DualConfigParser:
    """A dual config parser taking into account both system-wide files
    and user defaults. Any changes are stored in the user defaults."""

    def __init__(self, filename="timba.conf",
                 system_paths=_default_system_paths,
                 user_path=_default_user_path):
        self.syscp = ConfigParser()
        system_paths = [os.path.join(path, filename) for path in system_paths]
        self.syscp.read(system_paths)
        self.usercp = ConfigParser()
        self._user_file = os.path.join(user_path, "." + filename)
        self.usercp.read([self._user_file])

    def add_section(self, _section):
        if not self.syscp.has_section(_section):
            self.syscp.add_section(_section)
        if not self.usercp.has_section(_section):
            self.usercp.add_section(_section)

    def has_section(self, _section):
        return self.syscp.has_section(_section) or self.usercp.has_section(_section)

    def _get(self, _method, _section, _option, default=None, init=False, save=False):
        _section = _section or self.defsection  # TODO - defsection is unresolved for DualConfigParser
        # try user defaults
        try:
            return getattr(self.usercp, _method)(_section, _option)
        except (NoSectionError, NoOptionError, ValueError):
            error = sys.exc_info()[1]
        # try systemwide
        try:
            return getattr(self.syscp, _method)(_section, _option)
        except (NoSectionError, NoOptionError, ValueError):
            if default is not None and _option is not None:
                self.syscp.set(_section, _option, str(default))
                if init or save:
                    self.usercp.set(_section, _option, str(default))
                    if save:
                        self.usercp.write(open(self._user_file, "w"))
                return default
            # no default, so re-raise the error
            raise error

    def set(self, _section, _option, _value, save=True):
        _value = str(_value)
        # try to get option first, and do nothing if no change
        try:
            if self.get(_section, _option) == _value:
                return
        except (NoSectionError, NoOptionError):
            pass
        # save to user section
        try:
            self.usercp.add_section(_section)
        except DuplicateSectionError:
            pass
        self.usercp.set(_section, _option, _value)
        if save:
            with open(self._user_file, "w") as f:
                self.usercp.write(f)

    def has_option(self, _section, _option):
        return self.syscp.has_option(_section, _option) or \
               self.usercp.has_option(_section, _option)

    def get(self, _section, _option, default=None):
        return self._get('get', _option, default, _section)


class SectionParser:
    """A section parser is basically a ConfigParser with a default section name."""

    def __init__(self, _parser, _section):
        """Creates a SectionParser from a DualConfigParser and a section name"""
        self.parser = _parser
        _parser.add_section(_section)
        self.section = _section

    def has_option(self, _option, section=None):
        return self.parser.has_option(section or self.section, _option)

    def get(self, _option, default=None, section=None, init=False, save=False):
        return self.parser._get('get', section or self.section, _option, default, init=init, save=save)

    def getint(self, _option, default=None, section=None, init=False, save=False):
        return self.parser._get('getint', section or self.section, _option, default, init=init, save=save)

    def getfloat(self, _option, default=None, section=None, init=False, save=False):
        return self.parser._get('getfloat', section or self.section, _option, default, init=init, save=save)

    def getbool(self, _option, default=None, section=None, init=False, save=False):
        return self.parser._get('getboolean', section or self.section, _option, default, init=init, save=save)

    def set(self, _option, _value, section=None, save=True):
        return self.parser.set(section or self.section, _option, _value, save=save)


Config = DualConfigParser()
_section_parsers = {}


def section(name):
    global _section_parsers
    return _section_parsers.setdefault(name, SectionParser(Config, name))


if __name__ == '__main__':
    conf = Config('test')
    print('test1:', conf.get('test1', 1))
    print('test2:', conf.getint('test2', 2))
    print('test3:', conf.getfloat('test3', 3.0))
    try:
        print('test4:', conf.get('test4'))
    except:
        print('test4:', sys.exc_info())
    try:
        print('test5:', conf.get('test5'))
    except:
        print('test5:', sys.exc_info())
    conf.set('test6', 'abc')
    conf.set('test7', 1)
    conf.set('test8', 1.0)
    conf.set('test9', True)
    print('has test1:', conf.has_option('test1'))
    print('has test4:', conf.has_option('test4'))



"""Defines various useful functions and constants for parsing FITS headers"""

# Table of Stokes parameters corresponding to Stokes axis indices
# Taken from Table 7, Greisen, E. W., and Calabretta, M. R., Astronomy & Astrophysics, 395, 1061-1075, 2002
# (http://www.aanda.org/index.php?option=article&access=bibcode&bibcode=2002A%2526A...395.1061GFUL)
# So StokesNames[1] == "I", StokesNames[-1] == "RR", StokesNames[-8] == "YX", etc.
StokesNames = ["", "I", "Q", "U", "V", "YX", "XY", "YY", "XX", "LR", "RL", "LL", "RR"]
# complex axis convention
ComplexNames = ["", "real", "imag", "weight"]


def isAxisTypeX(ctype):
    """Checks if given CTYPE corresponds to the X axis"""
    return any([ctype.startswith(prefix) for prefix in ("RA", "GLON", "ELON", "HLON", "SLON")]) or \
           ctype in ("L", "X", "LL", "U", "UU")


def isAxisTypeY(ctype):
    """Checks if given CTYPE corresponds to the Y axis"""
    return any([ctype.startswith(prefix) for prefix in ("DEC", "GLAT", "ELAT", "HLAT", "SLAT")]) or \
           ctype in ("M", "Y", "MM", "V", "VV")



# Copyright (C) 2002-2022
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import math
import numpy as np
import operator

from PyQt5.QtWidgets import *
import traceback

from PyQt5.Qt import QObject, QHBoxLayout, QComboBox, QLabel, QLineEdit, QDialog, QVBoxLayout, Qt, QErrorMessage,\
    QSlider

import TigGUI.kitties.utils
from TigGUI.kitties.utils import curry
from TigGUI.kitties.widgets import BusyIndicator

QString = str

_verbosity = TigGUI.kitties.utils.verbosity(name="source_selector")
dprint = _verbosity.dprint
dprintf = _verbosity.dprintf

# list of standard tags to be sorted by
StandardTags = ["ra", "dec", "r", "Iapp", "I", "Q", "U", "V", "rm", "spi"]

# dict of accessors for nested source attributes
TagAccessors = dict()
for tag in "ra", "dec":
    TagAccessors[tag] = lambda src, t=tag: getattr(src.pos, t)
for tag in list("IQUV") + ["rm"]:
    TagAccessors[tag] = lambda src, t=tag: getattr(src.flux, t, 0.0)
for tag in ["spi"]:
    TagAccessors[tag] = lambda src, t=tag: getattr(src.spectrum, t, 0.0)

# tags for which sorting is not available
NonSortingTags = set(["name", "typecode"])


class SourceSelectorDialog(QDialog):
    def __init__(self, parent, flags=Qt.WindowFlags()):
        QDialog.__init__(self, parent, flags)
        self.setModal(False)
        self.setWindowTitle("Select sources by...")
        lo = QVBoxLayout(self)
        lo.setContentsMargins(10, 10, 10, 10)
        lo.setSpacing(5)
        # select by
        lo1 = QHBoxLayout()
        lo.addLayout(lo1)
        lo1.setContentsMargins(0, 0, 0, 0)
        #    lab = QLabel("Select:")
        #   lo1.addWidget(lab)
        self.wselby = QComboBox(self)
        lo1.addWidget(self.wselby, 0)
        self.wselby.activated[str].connect(self._setup_selection_by)
        # under/over
        self.wgele = QComboBox(self)
        lo1.addWidget(self.wgele, 0)
        self.wgele.addItems([">", ">=", "<=", "<", "sum<=", "sum>", "=="])
        self.wgele.activated[str].connect(self._select_threshold)
        # threshold value
        self.wthreshold = QLineEdit(self)
        self.wthreshold.editingFinished.connect(self._select_threshold)
        lo1.addWidget(self.wthreshold, 1)
        # min and max label
        self.wminmax = QLabel(self)
        lo.addWidget(self.wminmax)
        # selection slider
        lo1 = QHBoxLayout()
        lo.addLayout(lo1)
        self.wpercent = QSlider(self)
        self.wpercent.setTracking(False)
        self.wpercent.valueChanged[int].connect(self._select_percentile)
        self.wpercent.sliderMoved[int].connect(self._select_percentile_threshold)
        self.wpercent.setRange(0, 100)
        self.wpercent.setOrientation(Qt.Horizontal)
        lo1.addWidget(self.wpercent)
        self.wpercent_lbl = QLabel("0%", self)
        self.wpercent_lbl.setMinimumWidth(64)
        lo1.addWidget(self.wpercent_lbl)
        #    # hide button
        #    lo.addSpacing(10)
        #    lo2 = QHBoxLayout()
        #    lo.addLayout(lo2)
        #    lo2.setContentsMargins(0,0,0,0)
        #    hidebtn = QPushButton("Close",self)
        #    hidebtn.setMinimumWidth(128)
        #    QObject.connect(hidebtn,pyqtSignal("clicked()"),self.hide)
        #    lo2.addStretch(1)
        #    lo2.addWidget(hidebtn)
        #    lo2.addStretch(1)
        #    self.setMinimumWidth(384)
        self._in_select_threshold = False
        self._sort_index = None
        self.qerrmsg = QErrorMessage(self)

    def resetModel(self):
        """Resets dialog based on current model."""
        if not self.model:
            return
        # getset of model tags, and remove the non-sorting tags
        alltags = set(self.model.tagnames)
        alltags -= NonSortingTags
        # make list of tags from StandardTags that are present in model
        self.sorttags = [tag for tag in StandardTags if tag in alltags or tag in TagAccessors]
        # append model tags that were not in StandardTags
        self.sorttags += list(alltags - set(self.sorttags))
        # set selector
        self.wselby.clear()
        self.wselby.addItems(self.sorttags)
        for tag in "Iapp", "I":
            if tag in self.sorttags:
                self.wselby.setCurrentIndex(self.sorttags.index(tag))
                break
        self._setup_selection_by(self.wselby.currentText())

    def _reset_percentile(self):
        self.wthreshold.setText("")
        self.wpercent.setValue(50)
        self.wpercent_lbl.setText("--%")

    def _setup_selection_by(self, tag):
        tag = str(tag)  # may be QString
        # clear threshold value and percentiles
        self._reset_percentile()
        # get min/max values, and sort indices
        # _sort_index will be an array of (value,src,cumsum) tuples, sorted by tag value (high to low),
        # where src is the source, and cumsum is the sum of all values in the list from 0 up to and including the current one
        self._sort_index = []
        minval = maxval = value = None
        for isrc, src in enumerate(self.model.sources):
            try:
                if hasattr(src, tag):
                    # test if item can be cast to float
                    try:
                        float(getattr(src, tag))
                    except:
                        continue
                    else:
                        value = float(getattr(src, tag))

                elif tag in TagAccessors:
                    value = float(TagAccessors[tag](src))
                else: # not existant for this source (maybe a tag or something??)
                    value = np.nan
            # skip source if failed to access this tag as a float
            except:
                traceback.print_exc()
                continue
            if value is not None:
                self._sort_index.append([value if not np.isnan(value) else -np.inf, src, 0.])
                minval = min(minval, value if not np.isnan(value) else np.inf) if minval is not None else value
                maxval = max(maxval, value if not np.isnan(value) else -np.inf) if maxval is not None else value
        # add label
        if minval is None:
            self._range = None
            self.wminmax.setText("<font color=red>'%s' is not a numeric attribute</font>" % tag)
            for w in self.wgele, self.wthreshold, self.wpercent, self.wpercent_lbl:
                w.setEnabled(False)
        else:
            self._range = (minval, maxval)
            self.wminmax.setText("min: %g max: %g" % self._range)
            for w in self.wgele, self.wthreshold, self.wpercent, self.wpercent_lbl:
                w.setEnabled(True)
        # sort index by descending values
        self._sort_index.sort(reverse=True, key=operator.itemgetter(0))
        # generate cumulative sums
        cumsum = 0.
        for entry in self._sort_index:
            if not np.isneginf(entry[0]):
                cumsum += entry[0]
            entry[2] = cumsum

    # Maps comparison operators to callables. Used in _select_threshold.
    # Each callable takes two arguments: e is a tuple of (value,src,cumsum) (see _sort_index above), and x is a threshold
    # Second argument is a flag: if False, selection is inverted w.r.t. operator
    Operators = {
        "<": ((lambda e, x: e[0] >= x), False),
        "<=": ((lambda e, x: e[0] > x), False),
        ">": ((lambda e, x: e[0] > x), True),
        ">=": ((lambda e, x: e[0] >= x), True),
        "sum<=": ((lambda e, x: e[2] <= x), True),
        "sum>": ((lambda e, x: e[2] <= x), False),
        "==": ((lambda e, x: np.abs(e[0] - x) < 1.0e-8), True),
    }

    def _select_threshold(self, *dum):
        dprint(1, "select_threshold", dum)
        self._in_select_threshold = True
        busy = BusyIndicator()
        try:
            # get threshold, ignore if not set
            threshold = str(self.wthreshold.text())
            if not threshold:
                self._reset_percentile()
                return
            # try to parse threshold, ignore if invalid
            try:
                threshold = float(threshold)
            except:
                self._reset_percentile()
                return
            # get comparison operator
            op, select = self.Operators[str(self.wgele.currentText())]
            # apply to initial segment (that matches operator)
            for num, entry in enumerate(self._sort_index):
                if not op(entry, threshold):
                    break
                entry[1].selected = select
            else:
                num = len(self._sort_index)
            # apply to remaining segment
            for val, src, cumsum in self._sort_index[num:]:
                src.selected = not select
            # set percentile
            percent = round(float(num * 100) / len(self._sort_index))
            if not select:
                percent = 100 - percent
            self.wpercent.setValue(percent)
            self.wpercent_lbl.setText("%3d%%" % percent)
            # emit signal
            self.model.emitSelection(self)
        finally:
            self._in_select_threshold = False
            busy.reset_cursor()

    def _select_percentile(self, percent):
        self._select_percentile_threshold(percent, do_select=True)

    def _select_percentile_threshold(self, percent, do_select=False):
        # ignore if no sort index set up, or if _select_threshold() is being called
        if self._sort_index is None or self._in_select_threshold:
            return
        dprint(1, "select_precentile_threshold", percent)
        busy = BusyIndicator()
        # number of objects to select
        nsrc = len(self._sort_index)
        nsel = int(math.ceil(nsrc * float(percent) / 100))
        # get comparison operator
        opstr = str(self.wgele.currentText())
        op, select = self.Operators[opstr]
        # select head or tail of list, depending on direction of operator
        if select:
            thr = self._sort_index[min(nsel, nsrc - 1)]
            slc1 = slice(0, nsel)
            slc2 = slice(nsel, None)
        else:
            thr = self._sort_index[-min(nsel + 1, nsrc)]
            slc1 = slice(nsrc - nsel, None)
            slc2 = slice(0, nsrc - nsel)
        if do_select:
            for val, src, cumsum in self._sort_index[slc1]:
                src.selected = True
            for val, src, cumsum in self._sort_index[slc2]:
                src.selected = False
            self.model.emitSelection(self)
        self.wpercent_lbl.setText("%3d%%" % percent)
        self.wthreshold.setText("%g" % (thr[2] if opstr.startswith("sum") else thr[0]))
        busy.reset_cursor()
        return nsel

    def setModel(self, model):
        """Sets the current model. If dialog is visible, applies the changes"""
        self.model = model
        if self.isVisible():
            self.resetModel()
        if not model:
            self.hide()

    def show(self):
        """Shows dialog, resetting the model if it was invisible."""
        if not self.isVisible():
            self.resetModel()
        QDialog.show(self)


def show_source_selector(mainwin, model):
    dialog = getattr(mainwin, '_source_selector_dialog', None)
    if not dialog:
        dialog = mainwin._source_selector_dialog = SourceSelectorDialog(mainwin)
        mainwin.modelChanged.connect(dialog.setModel)
        mainwin.closing.connect(dialog.close)
    dialog.setModel(model)
    # show dialog
    dialog.show()
    dialog.raise_()

# from TigGUI.Tools import registerTool
# registerTool("Source selector...",show_source_selector)



# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from PyQt5 import QtWidgets
from PyQt5.Qt import QCursor, Qt, QWidgetAction, QLabel, QFrame, QTreeWidget, QObject, QApplication, \
    QTreeWidgetItemIterator, QListWidget
from PyQt5.QtWidgets import *
from PyQt5 import *
from PyQt5.QtCore import *


class BusyIndicator:
    """A BusyIndicator object is created to set the cursor to a hourglass.
    When the object is destroyed (i.e. when local variable goes out of scope), the cursor is reset."""

    def __init__(self):
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))

    def __del__(self):
        QApplication.restoreOverrideCursor()

    def reset_cursor(self):
        QApplication.restoreOverrideCursor()


def addMenuLabel(menu, text):
    """Adds a QLabel contaning text to the given menu"""
    qaw = QWidgetAction(menu)
    lab = QLabel(text, menu)
    qaw.setDefaultWidget(lab)
    lab.setAlignment(Qt.AlignCenter)
    lab.setFrameShape(QFrame.StyledPanel)
    lab.setFrameShadow(QFrame.Sunken)
    menu.addAction(qaw)
    return lab


class ClickableTreeWidget(QTreeWidget):

    def __init__(self, *args):
        QTreeWidget.__init__(self, *args)
        self._expanded_item = None
        self._mouse_press_pos = None
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested[QPoint].connect(self._request_context_menu)
        self.itemExpanded[QTreeWidgetItem].connect(self._item_expanded_collapsed)
        self.itemCollapsed[QTreeWidgetItem].connect(self._item_expanded_collapsed)

    def mousePressEvent(self, ev):
        self._expanded_item = None
        self._mouse_press_pos = ev.pos()
        QTreeWidget.mousePressEvent(self, ev)

    def mouseReleaseEvent(self, ev):
        item = self.itemAt(self._mouse_press_pos)
        col = None
        if item:
            col = self.header().logicalIndexAt(self._mouse_press_pos)
        # pass event to parent
        QTreeWidget.mouseReleaseEvent(self, ev)
        # now see if the item was expanded or collapsed because of the event. Only emit signal if this was
        # not the case (i.e. swallow the clicks that have to do with expansion/collapse of items)
        if item and item is not self._expanded_item and col is not None:
            self.itemClicked.emit(item, col)

    def _item_expanded_collapsed(self, item):
        self._expanded_item = item

    def _request_context_menu(self, pos):
        index = self.indexAt(pos)
        if not index.isValid():
            return

        item = self.itemAt(pos)
        if item:
            name = item.text(0)
            col = self.header().logicalIndexAt(pos)

            menu = QtWidgets.QMenu()
            menu.addSection("Menu")
            action = menu.addAction(name)
            action.setEnabled(False)
            menu.exec_(self.mapToGlobal(pos))

            # old line below fails
            # self.itemContextMenuRequested.emit(item, pos, col)

    class Iterator(QTreeWidgetItemIterator):
        def __init__(self, *args, **kw):
            QTreeWidgetItemIterator.__init__(self, *args)
            self._include_children = kw.get('children', False)
            parent = args[0]
            if isinstance(parent, QTreeWidget):
                if hasattr(QTreeWidget, 'invisibleRootItem'):
                    self._parent = parent.invisibleRootItem()
                else:
                    self._parent = None  # Qt 4.1 item.parent() returns None for top-level items
            else:
                self._parent = parent

        def __iter__(self):
            return self

        def __next__(self):
            while True:
                value = self.value()
                self.__iadd__(1)
                if not value:
                    raise StopIteration
                if self._include_children or value.parent() is None or value.parent() is self._parent:
                    return value

        def next(self):
            return self.__next__()

    def iterator(self, *args):
        """Returns a child item iterator.
        iterator([flags]) returns an iterator for the tree widget itself
        iterator(item,[flags]) returns an iterator for a tree widget item
        """
        if len(args) > 1:
            return ClickableTreeWidget.Iterator(*args)
        else:
            return ClickableTreeWidget.Iterator(self, *args)


TreeWidgetItemIterator = ClickableTreeWidget.Iterator


class ClickableListWidget(QListWidget):
    def __init__(self, *args):
        QListWidget.__init__(self, *args)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested[QPoint].connect(self._request_context_menu)

    def _request_context_menu(self, pos):
        item = self.itemAt(pos)
        if item:
            self.itemContextMenuRequested.emit(item, pos)



# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import math

import numpy
import numpy.ma
import os.path
import time

from PyQt5 import QtCore
from PyQt5.Qt import QObject, QRect, QRectF, QPointF, QPoint, QSizeF
from PyQt5.Qwt import QwtPlotItem
from PyQt5.QtCore import pyqtSignal
from scipy.ndimage import interpolation, measurements

import TigGUI.kitties.utils

from astropy.io import fits as pyfits

from Tigger.Coordinates import Projection
from TigGUI.Images import Colormaps
from Tigger.Tools import FITSHeaders


DEG = math.pi / 180

_verbosity = TigGUI.kitties.utils.verbosity(name="skyimage")
dprint = _verbosity.dprint
dprintf = _verbosity.dprintf


class SkyImagePlotItem(QwtPlotItem, QObject):
    """SkyImagePlotItem is a 2D image in l,m coordimnates"""

    def __init__(self, nx=0, ny=0, l0=0, m0=0, dl=1, dm=1, image=None):
        QwtPlotItem.__init__(self)
        # name, if any
        self.name = self.filename = None
        # internal init
        self.RenderAntialiased
        self._qo = QObject()
        self.qimg = None
        self._image = self._imgminmax = None
        self._nvaluecalls = 0
        self._value_time = self._value_time0 = None
        self._lminmax = (0, 0)
        self._mminmax = (0, 0)
        self._cache_qimage = {}
        self._cache_mapping = self._cache_imap = self._cache_interp = None
        self._psfsize = 0, 0, 0
        #self.projection = None
        self._nx, self._ny = 0, 0
        self._l0, self._m0 = 0, 0
        self._dl, self._dm = 0, 0
        self._x0, self._y0 = 0, 0
        self._lminmax = None
        self._mminmax = None
        self._bounding_rect = None
        self._bounding_rect_pix = None
        self._image_key = None
        self._prefilter = None
        self._current_rect = None
        self._current_rect_pix = None
        # set image, if specified
        if image is not None:
            nx, ny = image.shape
            self.setImage(image)
        # set coordinates, if specified
        if nx and ny:
            self.setImageCoordinates(nx, ny, l0, m0, dl, dm)
        # set default colormap and intensity map
        self.colormap = Colormaps.GreyscaleColormap
        self.imap = Colormaps.LinearIntensityMap()
        self.signalRepaint = None
        self.signalSlice = None
        self.signalRaise = None
        self.signalUnload = None
        self.signalCenter = None
        self.signalPlotRaised = None

    # start of signal connection methods

    def connectRepaint(self, sig):
        self.signalRepaint = sig

    def connectSlice(self, sig):
        self.signalSlice = sig

    def connectRaise(self, sig):
        self.signalRaise = sig

    def connectUnload(self, sig):
        self.signalUnload = sig

    def connectCenter(self, sig):
        self.signalCenter = sig

    def connectPlotRiased(self, sig):
        self.signalPlotRaised = sig

    # end  of signal connection methods

    def clearDisplayCache(self):
        """Clears all display caches."""
        self._cache_qimage = {}
        self._cache_interp = self._cache_imap = None

    def setColorMap(self, cmap=None, emit=True):
        """Changes the colormap. If called with no arguments, clears colormap-dependent caches"""
        self._cache_qimage = {}
        if cmap:
            self.colormap = cmap
        if emit:
            self.signalRepaint.emit()

    def updateCurrentColorMap(self):
        self._cache_qimage = {}
        self.signalRepaint.emit()

    def setIntensityMap(self, imap=None, emit=True):
        """Changes the intensity map. If called with no arguments, clears intensity map-dependent caches"""
        self._cache_qimage = {}
        self._cache_imap = None
        if imap:
            self.imap = imap
        if emit:
            self.signalRepaint.emit()

    def colorMap(self):
        return self.colormap

    def intensityMap(self):
        return self.imap

    def setImageCoordinates(self, nx, ny, x0, y0, l0, m0, dl, dm):
        """Sets up image coordinates. Pixel x0,y0 is centered at location l0,m0 in the plot, pixel size is dl,dm, image size is (nx,ny)"""
        dprint(2, "image coordinates are", nx, ny, x0, y0, l0, m0, dl, dm)
        self._nx, self._ny = nx, ny
        self._l0, self._m0 = l0, m0
        self._dl, self._dm = dl, dm
        self._x0, self._y0 = x0, y0
        self._lminmax = (l0 - dl * (x0 + 0.5), l0 + (nx - x0 - 0.5) * dl)
        if dl < 0:
            self._lminmax = (self._lminmax[1], self._lminmax[0])
        self._mminmax = (m0 - dm * (y0 + 0.5), m0 + (ny - y0 - 0.5) * dm)
        self._bounding_rect = QRectF(self._lminmax[0], self._mminmax[0], nx * abs(dl), ny * abs(dm))
        self._bounding_rect_pix = QRect(0, 0, nx, ny)
        dprint(2, "image extents are", self._lminmax, self._mminmax)

    def imageDims(self):
        """Returns image dimensions as mx,ny"""
        return self._nx, self._ny

    def referencePixel(self):
        return self._x0, self._y0

    def lmToPix(self, l, m):
        """Converts l,m coordimnates to float (so possibly fractional) pixel coordinates."""
        return self._x0 + (l - self._l0) / self._dl, self._y0 + (m - self._m0) / self._dm

    def pixToLm(self, x, y):
        """Converts pixel coordinates to lm coordinates."""
        return self._l0 + (x - self._x0) * self._dl, self._m0 + (y - self._y0) * self._dm

    def getExtents(self):
        """Returns image extent, as (l0,l1),(m0,m1)"""
        return self._lminmax, self._mminmax

    def boundingRect(self):
        """Returns bouding rectangle of image, in lm coordinates."""
        return self._bounding_rect

    def currentRect(self):
        """Returns currently visible rectange, in lm coordinates. Coordinates may be outside of image range."""
        return self._current_rect

    def currentRectPix(self):
        """Returns currently visible rectange, in pixel coordinates. Pixel coordinates are bounded to 0,0 and nx-1,ny-1."""
        return self._current_rect_pix

    def setImage(self, image, key=None, minmax=None):
        """Sets image array.
        If key is not None, sets this as the image key (for use with the pixmap cache.)
        If minmax is not None, then stores this as the (presumably cached or precomputed) min/max values.
        """
        self._image = image
        self._imgminmax = minmax
        self._image_key = key
        # clear intermediate caches
        self._prefilter = self._cache_interp = self._cache_imap = None
        # if key is None, also clear QImage cache -- it only works when we have images identified by keys
        if key is None:
            self._cache_qimage = {}

    def image(self):
        """Returns image array."""
        return self._image

    def imagePixel(self, x, y):
        if numpy.ma.isMA(self._image):
            return self._image.data[x, y], self._image.mask[x, y]
        else:
            return self._image[x, y], False

    def imageMinMax(self):
        if not self._imgminmax:
            dprint(3, "computing image min/max")
            rdata, rmask = self.optimalRavel(self._image)
            try:
                self._imgminmax = measurements.extrema(rdata, labels=rmask, index=None if rmask is None else False)[:2]
            except:
                # when all data is masked, some versions of extrema() throw an exception
                self._imgminmax = numpy.nan, numpy.nan
            dprint(3, self._imgminmax)
        return self._imgminmax

    def draw(self, painter, xmap, ymap, rect, use_cache=True):
        """Implements QwtPlotItem.draw(), to render the image on the given painter."""
        xp1, xp2, xdp, xs1, xs2, xds = xinfo = xmap.p1(), xmap.p2(), xmap.pDist(), xmap.s1(), xmap.s2(), xmap.sDist()
        yp1, yp2, ydp, ys1, ys2, yds = yinfo = ymap.p1(), ymap.p2(), ymap.pDist(), ymap.s1(), ymap.s2(), ymap.sDist()
        dprint(5, "draw:", rect, xinfo, yinfo)
        self._current_rect = QRectF(QPointF(xs2, ys1), QSizeF(xds, yds))
        self._current_rect_pix = QRectF(QPointF(*self.lmToPix(xs1, ys1)), QPointF(*self.lmToPix(xs2, ys2))).toRect().intersected(
            self._bounding_rect_pix)
        dprint(5, "draw:", self._current_rect_pix)
        # put together tuple describing current mapping
        mapping = xinfo, yinfo
        # if mapping has changed w.r.t. cache (i.e. zoom has changed), discard all cached QImages
        if mapping != self._cache_mapping:
            dprint(2, "does not match cached mapping, cache is:", self._cache_mapping)
            dprint(2, "and we have:", mapping)
            self.clearDisplayCache()
            self._cache_mapping = mapping
        t0 = time.time()
        # check cached QImage for current image key.
        self.qimg = self._cache_qimage.get(self._image_key)
        if self.qimg:
            dprint(5, "QImage found in cache, reusing")
        # else regenerate image
        else:
            # check for cached intensity-mapped data
            if self._cache_imap is not None:
                dprint(5, "intensity-mapped data found in cache, reusing")
            else:
                if self._cache_interp is not None:
                    dprint(5, "interpolated data found in cache, reusing")
                else:
                    image = self._image.transpose() if self._data_fortran_order else self._image
                    spline_order = 2
                    xsamp = abs(xmap.sDist() / xmap.pDist()) / abs(self._dl)
                    ysamp = abs(ymap.sDist() / ymap.pDist()) / abs(self._dm)
                    if max(xsamp, ysamp) < .33 or min(xsamp, ysamp) > 2:
                        spline_order = 1
                    dprint(2, "regenerating drawing cache, sampling factors are", xsamp, ysamp, "spline order is",
                           spline_order)
                    self._cache_imap = None
                    if self._prefilter is None and spline_order > 1:
                        self._prefilter = interpolation.spline_filter(image, order=spline_order)
                        dprint(2, "spline prefiltering took", time.time() - t0, "secs")
                        t0 = time.time()
                    # make arrays of plot coordinates
                    # xp[0],yp[0] corresponds to pixel 0,0, where 0,0 is the upper-left corner of the plot
                    # the maps are in a funny order (w.r.t. meaning of p1/p2/s1/s2), so the indices here are determined empirically
                    # We also adjust by half-pixel, to get the world coordinate of the pixel _center_
                    xp = xmap.s1() - (xmap.sDist() / xmap.pDist()) * (0.5 + numpy.arange(int(xmap.pDist())))
                    yp = ymap.s2() - (ymap.sDist() / ymap.pDist()) * (0.5 + numpy.arange(int(ymap.pDist())))
                    # now convert plot coordinates into fractional image pixel coordinates
                    xi = self._x0 + (xp - self._l0) / self._dl
                    yi = self._y0 + (yp - self._m0) / self._dm
                    # interpolate image data
                    ###        # old code for nearest-neighbour interpolation
                    ###        # superceded by interpolation below (we simply round pixel coordinates to go to NN when oversampling)
                    ###        xi = xi.round().astype(int)
                    ###        oob_x = (xi<0)|(xi>=self._nx)
                    ###        xi[oob_x] = 0
                    ###        yi = yi.round().astype(int)
                    ###        oob_y = (yi<0)|(yi>=self._ny)
                    ###        yi[oob_y] = 0
                    ###        idx = (xi[:,numpy.newaxis]*self._ny + yi[numpy.newaxis,:]).ravel()
                    ###        interp_image = self._image.ravel()[idx].reshape((len(xi),len(yi)))
                    ###        interp_image[oob_x,:] = 0
                    ###        interp_image[:,oob_y] = 0
                    ###        self._qimage_cache = self.colormap.colorize(interp_image,self._img_range)
                    ###        self._qimage_cache_attrs = (rect,xinfo,yinfo)

                    # if either axis is oversampled by a factor of 3 or more, switch to nearest-neighbour interpolation by rounding pixel values
                    if xsamp < .33:
                        xi = xi.round()
                    if ysamp < .33:
                        yi = yi.round()
                    # make [2,nx,ny] array of interpolation coordinates
                    xy = numpy.zeros((2, len(xi), len(yi)))
                    xy[0, :, :] = xi[:, numpy.newaxis]
                    xy[1, :, :] = yi[numpy.newaxis, :]
                    # interpolate. Use NAN for out of range pixels...
                    # for fortran order, tranpose axes for extra speed (flip XY around then)
                    if self._data_fortran_order:
                        xy = xy[-1::-1, ...]
                    if spline_order > 1:
                        interp_image = interpolation.map_coordinates(self._prefilter, xy, order=spline_order,
                                                                     cval=numpy.nan, prefilter=False)
                    else:
                        interp_image = interpolation.map_coordinates(image, xy, order=spline_order, cval=numpy.nan)
                    # ...and put a mask on them (Colormap.colorize() will make these transparent).
                    mask = ~numpy.isfinite(interp_image)
                    self._cache_interp = numpy.ma.masked_array(interp_image, mask)
                    dprint(2, "interpolation took", time.time() - t0, "secs")
                    t0 = time.time()
                # ok, we have interpolated data in _cache_interp
                self._cache_imap = self.imap.remap(self._cache_interp)
                dprint(2, "intensity mapping took", time.time() - t0, "secs")
                t0 = time.time()
            # ok, we have intensity-mapped data in _cache_imap
            self.qimg = self.colormap.colorize(self._cache_imap)
            dprint(2, "colorizing took", time.time() - t0, "secs")
            t0 = time.time()
            if use_cache:
                # cache the qimage
                self._cache_qimage[self._image_key] = self.qimg.copy()
        # now draw the image
        t0 = time.time()
        painter.drawImage(QPointF(xp1, yp2), self.qimg)
        dprint(2, "drawing took", time.time() - t0, "secs")
        # when exporting images to PNG cache needs to be cleared
        if not use_cache:
            self.clearDisplayCache()

    def setPsfSize(self, _maj, _min, _pa):
        self._psfsize = _maj, _min, _pa

    def getPsfSize(self):
        return self._psfsize


ScalePrefixes = ["p", "n", "\u03bc", "m", "", "K", "M", "G", "T"]


def getScalePrefix(*values):
    """Helper method to get the optimal scale and SI prefix for a given range of values"""
    # take log10. If all values are zero, use prefix of 1.
    log10 = numpy.ma.log10(numpy.ma.abs(values))
    if log10.mask.all():
        return 1, ""
    # find appropriate prefix
    # Add 1 to log10(min) (so that >=.1 unit is reported as unit), divide by 3, take floor, look up unit prefix
    m = int(math.floor((log10.min() + 1) / 3)) + 4
    m = max(m, 0)
    m = min(m, len(ScalePrefixes) - 1)
    return 10 ** ((m - 4) * 3), ScalePrefixes[m]


class SkyCubePlotItem(SkyImagePlotItem):
    """Extends SkyImagePlotItem with a hypercube containing extra slices."""

    def __init__(self, data=None, ndim=None):
        SkyImagePlotItem.__init__(self)
        self.RenderAntialiased
        # datacube (array of any rank)
        self._data_fortran_order = None
        self._data = self._dataminmax = None
        # current image slice (a list of indices) applied to data to make an image
        self.imgslice = None
        # info about sky axes
        self._skyaxes = [None, None]
        # info about other axes
        self._extra_axes = []
        # set other info
        self.ra0 = 0
        self.dec0 = 0
        if data is not None:
            self.setData(data)
        elif ndim:
            self.setNumAxes(ndim)

    def setData(self, data, fortran_order=False):
        """Sets the datacube. fortran_order is a hint, which makes iteration over
        fortran-order arrays faster when computing min/max and such."""
        # Note that iteration order is absolutely critical for large cubes -- if data is in fortran
        # order in memory, then that's the way we should iterate over it, period. Transposing is too
        # slow. We therefore create 1D "views" of the data using numpy.ravel(x,order='F'), and use
        # thse to iterate over the data for things like min/max, masking, etc.
        if fortran_order:
            dprint(3, "setData: computing mask (fortran order)")
            rav = numpy.ravel(data, order='F')
            rfin = numpy.isfinite(rav)
            if rfin.all():
                dprint(3, "setData: phew, all finite, nothing to be masked")
                self._data = data
            else:
                dprint(3, "setData: setting masked elements to 0")
                rmask = ~rfin
                rav[rmask] = 0
                dprint(3, "setData: creating masked array")
                mask = rmask.reshape(data.shape[-1::-1]).transpose()
                self._data = numpy.ma.masked_array(data, mask)
        else:
            dprint(3, "setData: computing mask (C order)")
            fin = numpy.isfinite(data)
            if fin.all():
                dprint(3, "setData: phew, all finite, nothing to be masked")
                self._data = data
            else:
                dprint(3, "setData: setting masked elements to 0")
                mask = ~fin
                data[mask] = 0
                dprint(3, "setData: creating masked array")
                self._data = numpy.ma.masked_array(data, mask)
        dprint(3, "setData: wrapping up")
        self._data_fortran_order = fortran_order
        self._dataminmax = None
        self.setNumAxes(data.ndim)
        ### old slow code
        # dprint(3,"setData: computing mask")
        # fin = numpy.isfinite(data)
        # mask = ~fin
        # dprint(3,"setData: setting masked elements to 0")
        # data[mask] = 0
        # dprint(3,"setData: creating masked array")
        # self._data = numpy.ma.masked_array(data,mask)
        # dprint(3,"setData: wrapping up")
        # self._data_fortran_order = fortran_order
        # self._dataminmax = None
        # self.setNumAxes(data.ndim)

    def data(self):
        """Returns datacube"""
        return self._data

    def isDataInFortranOrder(self):
        return self._data_fortran_order

    def optimalRavel(self, array):
        """Returns the "optimal ravel" corresponding to the given array, which is either FORTRAN
        or C order. The optimal ravel is that over which iteration is fastest.
        Returns tuple of ravarray,ravmask. If input array is not masked, then ravmask=None."""
        order = 'F' if self._data_fortran_order else 'C'
        rarr = numpy.ravel(array, order=order)
        rmask = numpy.ravel(array.mask, order=order) if numpy.ma.isMA(array) else None
        return rarr, rmask

    def dataMinMax(self):
        if not self._dataminmax:
            rdata, rmask = self.optimalRavel(self._data)
            dprint(3, "computing data min/max")
            try:
                self._dataminmax = measurements.extrema(rdata, labels=rmask, index=None if rmask is None else False)
            except:
                # when all data is masked, some versions of extrema() throw an exception
                self._dataminmax = numpy.nan, numpy.nan
            dprint(3, self._dataminmax)
        return self._dataminmax

    def setNumAxes(self, ndim):
        self.imgslice = [0] * ndim

    def setSkyAxis(self, n, iaxis, nx, x0, dx, xpix0):
        """Sets the sky axis, n=0 for RA and n=1 for Dec"""
        if not self.imgslice:
            raise RuntimeError("setNumAxes() must be called first")
        # reverse axis if step is negative
        #    if dx<0:
        #      dx = -dx
        #      xpix0 = nx-1-xpix0
        #      self.imgslice[iaxis] = slice(-1,None,-1)
        #    else:
        self.imgslice[iaxis] = slice(None)
        self._skyaxes[n] = iaxis, nx, x0, dx, xpix0
        if iaxis == 0:
            self.ra0 = x0
        else:
            self.dec0 = x0

    def getSkyAxis(self, n):
        return self._skyaxes[n][:2]

    def setExtraAxis(self, iaxis, name, labels, values, units):
        """Sets additional hypercube axis. labels is an array of strings, one per each axis element, for labelled axes, or None if axis should be labelled with values/units.
        values is an array of axis values, and units are the units in which values are expressed.
        """
        units = units or ""
        scale, prefix = getScalePrefix(values)
        units = prefix + units
        # estimate number of significant digits
        valarr = numpy.array(values) / scale
        try:
            ndigits = int(math.ceil(math.log10(max(abs(valarr)) / abs((valarr[1:] - valarr[0:-1])).min())))
            nexp = int(abs(numpy.log10(abs(valarr))).max())
            #     print ndigits,nexp
            if nexp > 4:
                format = ".%de" % ndigits
            else:
                format = ".%df" % ndigits
        except:
            format = ".2g"
        if labels is None:
            labels = [("%d: %" + format + " %s") % (i, val / scale, units) for i, val in enumerate(values)]
        self._extra_axes.append((iaxis, name, labels, values, units, scale))

    def numExtraAxes(self):
        return len(self._extra_axes)

    def extraAxisNumberNameLabels(self, i):
        return self._extra_axes[i][:3]

    def extraAxisValues(self, i):
        return self._extra_axes[i][3]

    def extraAxisUnitScale(self, i):
        return self._extra_axes[i][4:6]

    def setPlotProjection(self, proj=None):
        """Sets the projection of the plot. Must be called before image is drawn. If None is given, the default
        projection is used.
        """
        if not (self._skyaxes[0] and self._skyaxes[1]):
            raise RuntimeError("setSkyAxis() must be called for both sky axes")
        (iaxis_ra, nx, ra0, dra, i0), (iaxis_dec, ny, dec0, ddec, j0) = self._skyaxes
        proj = proj or self.projection
        # setup projection properties and get center of field
        l0, m0 = proj.lm(ra0, dec0)
        # find cell sizes
        if proj is self.projection:
            dl, dm = -self.projection.xscale, self.projection.yscale
        else:
            dl = proj.offset(dra, 0)[0]
            dm = proj.offset(0, ddec)[1]
        # setup image coordinates
        self.setImageCoordinates(nx, ny, i0, j0, l0, m0, dl, dm)

    def setDefaultProjection(self, projection=None):
        """Sets default image projection. If None is given, sets up default SinWCS projection."""
        # FITSWCS_static does not seem to be called often, if at all.
        self.projection = projection or Projection.FITSWCS_static(self.ra0, self.dec0)
        self.setPlotProjection()

    def _setupSlice(self):
        index = tuple(self.imgslice)
        key = tuple([index[iaxis] for iaxis, name, labels, values, units, scale in self._extra_axes])
        image = self._data[index]
        self.setImage(self._data[index], key=key)

    def selectSlice(self, *indices):
        if len(indices) != len(self._extra_axes):
            raise ValueError("number of indices does not match number of extra axes""")
        for i, (iaxis, name, labels, values, units, scale) in enumerate(self._extra_axes):
            self.imgslice[iaxis] = indices[i]
        self._setupSlice()
        self.signalSlice.emit(indices)

    def currentSlice(self):
        return list(self.imgslice)


class FITSImagePlotItem(SkyCubePlotItem):
    fits_header = None

    @staticmethod
    def hasComplexAxis(hdr):
        """Returns True if given FITS header has a complex axis (must be last axis)"""
        nax = hdr['NAXIS']
        return nax if hdr['CTYPE%d' % nax].strip() == "COMPLEX" else 0

    @staticmethod
    def addComplexAxis(header):
        """Adds a complex axis to the given FITS header, returns new copy of header"""
        hdr = header.copy()
        nax = hdr['NAXIS'] + 1
        hdr['NAXIS'] = nax
        hdr.set('NAXIS%d' % nax, 2, "complex image")
        hdr.set('CTYPE%d' % nax, "COMPLEX", "complex image")
        hdr.set('CRPIX%d' % nax, 1)
        hdr.set('CRVAL%d' % nax, 1)
        hdr.set('CDELT%d' % nax, 1)
        return hdr

    @staticmethod
    def removeComplexAxis(header):
        """Removes a complex axis from the given FITS header, returns new copy of header"""
        axis = FITSImagePlotItem.hasComplexAxis(header)
        if axis:
            header = header.copy()
            header['NAXIS'] = axis - 1
            for name in 'NAXIS', 'CTYPE', 'CRPIX', 'CRVAL', 'CDELT':
                key = "%s%d" % (name, axis)
                if key in header:
                    del header[key]
        return header

    def __init__(self, filename=None, name=None, hdu=None):
        SkyCubePlotItem.__init__(self)
        self.RenderAntialiased
        self.name = name
        if filename or hdu:
            self.read(filename, hdu)

    StokesNames = FITSHeaders.StokesNames
    ComplexNames = FITSHeaders.ComplexNames

    def read(self, filename, hdu=None):
        self.filename = filename
        self.name = self.name or os.path.basename(filename)
        # read FITS file
        if not hdu:
            dprint(3, "opening", filename)
            hdu = pyfits.open(filename)[0]
            hdu.verify('silentfix')
            if os.path.getsize(filename) < hdu._file.tell():
                raise RuntimeError(
                    f"FITS file may have been truncated: file length ({os.path.getsize(filename)}) "
                    f"is smaller than expected ({hdu._file.tell()})")
        hdr = self.fits_header = hdu.header
        dprint(3, "reading data")
        data = hdu.data
        # NB: all-data operations (such as getting global min/max or computing of histograms) are much faster
        # (almost x2) when data is iterated
        # over in the proper order. After a transpose(), data is in fortran order. Tell this to setData().
        data = numpy.transpose(data)  # .copy()
        dprint(3, "setting data")
        self.setData(data, fortran_order=True)
        dprint(3, "reading header")
        ndim = hdr['NAXIS']
        if ndim < 2:
            raise ValueError("Cannot load a one-dimensional FITS file")
        # setup projection
        # (strip out history from header, as big histories really slow down FITSWCS)
        hdr1 = pyfits.Header([x for x in hdr.cards if not str(x).startswith('HISTORY')])
        proj = Projection.FITSWCS(hdr1)
        nx = ny = None
        # find X and Y axes
        for iaxis in range(ndim):
            axs = str(iaxis + 1)
            npix = hdr['NAXIS' + axs]
            name = hdr.get('CTYPE' + axs, axs).strip().upper()
            # have we found the coordinate axes?
            if FITSHeaders.isAxisTypeX(name):
                nx = npix
                iaxis_ra = iaxis
            elif FITSHeaders.isAxisTypeY(name):
                ny = npix
                iaxis_dec = iaxis
        # check that we have them
        if nx is None or ny is None:
            iaxis_ra, iaxis_dec = 0, 1
            nx, ny = hdr.get('NAXIS1'), hdr.get('NAXIS2')
        for iaxis in range(ndim):
            axs = str(iaxis + 1)
            # get axis description
            npix = hdr['NAXIS' + axs]
            crval = hdr.get('CRVAL' + axs, 0)
            cdelt = hdr.get('CDELT' + axs, 1)
            crpix = hdr.get('CRPIX' + axs, 1) - 1
            name = hdr.get('CTYPE' + axs, axs).strip().upper()
            unit = hdr.get('CUNIT' + axs)
            # if this is not an X/Y axis, add it to the slicers
            if iaxis not in (iaxis_ra, iaxis_dec):
                # values becomes a list of axis values
                values = list(crval + (numpy.arange(npix) - crpix) * cdelt)
                unit = unit and unit.lower().capitalize()
                # FITS knows of two enumerable axes: STOKES and COMPLEX. For these two, replace values with proper names
                if name == "STOKES":
                    labels = [(self.StokesNames[int(i)] if i > 0 and i < len(self.StokesNames) else "%d" % i) for i in
                              values]
                elif name == "COMPLEX":
                    labels = [(self.ComplexNames[int(i)] if i > 0 and i < len(self.ComplexNames) else "%d" % i) for i in
                              values]
                else:
                    name = name.split("-")[0]
                    # if values are a simple sequence startying at 0 or 1, make simple labels
                    if cdelt == 1 and values[0] in (0., 1.):
                        labels = ["%d%s" % (val, unit) for val in values]
                    # else set labels to None: setExtraAxis() will figure it out
                    else:
                        labels = None
                self.setExtraAxis(iaxis, name or ("axis " + axs), labels, values, unit)
        # check for beam parameters
        psf = [hdr.get(x, None) for x in ('BMAJ', 'BMIN', 'BPA')]
        if all([x is not None for x in psf]):
            self.setPsfSize(*[p / 180 * math.pi for p in psf])
        self.setSkyAxis(0, iaxis_ra, nx, proj.ra0, -proj.xscale, proj.xpix0)
        self.setSkyAxis(1, iaxis_dec, ny, proj.dec0, proj.yscale, proj.ypix0)
        self.setDefaultProjection(proj)
        dprint(3, "setting initial slice")
        self._setupSlice()

    def save(self, filename):
        data = data1 = self.data().transpose()
        if numpy.ma.isMA(data):
            data1 = data.data.copy()
            data1[data.mask] = numpy.NAN
        hdu = pyfits.PrimaryHDU(data1, self.fits_header)
        hdu.verify('silentfix')
        if os.path.exists(filename):
            os.remove(filename)
        hdu.writeto(filename, overwrite=True)
        self.filename = filename
        self.name = os.path.basename(filename)



# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import math
from PyQt5.QtWidgets import *

from PyQt5.Qt import QObject, QHBoxLayout, QFileDialog, pyqtSignal, QLabel, \
    QLineEdit, QDialog, QDoubleValidator, QVBoxLayout, \
    QPushButton, Qt, QGridLayout, QMessageBox, QErrorMessage

import TigGUI.kitties.utils

from astropy.io import fits as pyfits

import os.path

from TigGUI.kitties.widgets import BusyIndicator
from TigGUI.Widgets import FileSelector
from Tigger.Models import SkyModel, ModelClasses

DEG = math.pi / 180

from astLib.astWCS import WCS


class AddBrickDialog(QDialog):
    def __init__(self, parent, modal=True, flags=Qt.WindowFlags()):
        QDialog.__init__(self, parent, flags)
        self.model = None
        self._model_dir = None
        self.setModal(modal)
        self.setWindowTitle("Add FITS brick")
        lo = QVBoxLayout(self)
        lo.setContentsMargins(10, 10, 10, 10)
        lo.setSpacing(5)
        # file selector
        self.wfile = FileSelector(self, label="FITS filename:", dialog_label="FITS file", default_suffix="fits",
                                  file_types="FITS files (*.fits *.FITS)", file_mode=QFileDialog.ExistingFile)
        lo.addWidget(self.wfile)
        # overwrite or add mode
        lo1 = QGridLayout()
        lo.addLayout(lo1)
        lo1.setContentsMargins(0, 0, 0, 0)
        lo1.addWidget(QLabel("Padding factor:", self), 0, 0)
        self.wpad = QLineEdit("2", self)
        self.wpad.setValidator(QDoubleValidator(self))
        lo1.addWidget(self.wpad, 0, 1)
        lo1.addWidget(QLabel("Assign source name:", self), 1, 0)
        self.wname = QLineEdit(self)
        lo1.addWidget(self.wname, 1, 1)
        # OK/cancel buttons
        lo.addSpacing(10)
        lo2 = QHBoxLayout()
        lo.addLayout(lo2)
        lo2.setContentsMargins(0, 0, 0, 0)
        lo2.setContentsMargins(5, 5, 5, 5)
        self.wokbtn = QPushButton("OK", self)
        self.wokbtn.setMinimumWidth(128)
        self.wokbtn.clicked.connect(self.accept)
        self.wokbtn.setEnabled(False)
        cancelbtn = QPushButton("Cancel", self)
        cancelbtn.setMinimumWidth(128)
        cancelbtn.clicked.connect(self.reject)
        lo2.addWidget(self.wokbtn)
        lo2.addStretch(1)
        lo2.addWidget(cancelbtn)
        self.setMinimumWidth(384)
        # signals
        self.wfile.filenameSelected.connect(self._fileSelected)
        # internal state
        self.qerrmsg = QErrorMessage(self)

    def setModel(self, model):
        self.model = model
        if model.filename():
            self._model_dir = os.path.dirname(os.path.abspath(model.filename()))
        else:
            self._model_dir = os.path.abspath('.')
        self.wfile.setDirectory(self._model_dir)
        self._fileSelected(self.wfile.filename(), quiet=True)

    def _fileSelected(self, filename, quiet=False):
        self.wokbtn.setEnabled(False)
        if not filename:
            return None
        # check that filename matches model
        if not os.path.samefile(self._model_dir, os.path.dirname(filename)):
            self.wfile.setFilename('')
            if not quiet:
                QMessageBox.warning(self, "Directory mismatch", """<P>The FITS file must reside in the same directory
          as the current sky model.</P>""")
            self.wfile.setDirectory(self._model_dir)
            return None
        # if filename is not in model already, enable the "add to model" control
        for src in self.model.sources:
            if isinstance(getattr(src, 'shape', None), ModelClasses.FITSImage):
                if os.path.exists(src.shape.filename) and os.path.samefile(src.shape.filename, filename):
                    if not quiet:
                        QMessageBox.warning(self, "Already in model",
                                            "This FITS brick is already present in the model.")
                    self.wfile.setFilename('')
                    return None
        if not str(self.wname.text()):
            self.wname.setText(os.path.splitext(os.path.basename(str(filename)))[0])
        self.wokbtn.setEnabled(True)
        return filename

    def accept(self):
        """Tries to add brick, and closes the dialog if successful."""
        filename = self.wfile.filename()
        # read fits file
        busy = BusyIndicator()
        try:
            input_hdu = pyfits.open(filename)[0]
        except Exception as err:
            busy.reset_cursor()
            QMessageBox.warning(self, "Error reading FITS", "Error reading FITS file %s: %s" % (filename, str(err)))
            return
        # check name
        srcname = str(self.wname.text()) or os.path.splitext(os.path.basename(str(filename)))[0]
        if srcname in set([src.name for src in self.model.sources]):
            QMessageBox.warning(self, "Already in model",
                                "<p>The model already contains a source named '%s'. Please select a different name.</p>" % srcname)
            return
        # get image parameters
        hdr = input_hdu.header
        max_flux = float(input_hdu.data.max())
        wcs = WCS(hdr, mode='pyfits')
        # Get reference pixel coordinates
        # wcs.getCentreWCSCoords() doesn't work, as that gives us the middle of the image
        # So scan the header to get the CRPIX values
        ra0 = dec0 = 1
        for iaxis in range(hdr['NAXIS']):
            axs = str(iaxis + 1)
            name = hdr.get('CTYPE' + axs, axs).upper()
            if name.startswith("RA"):
                ra0 = hdr.get('CRPIX' + axs, 1) - 1
            elif name.startswith("DEC"):
                dec0 = hdr.get('CRPIX' + axs, 1) - 1
        # convert pixel to degrees
        #    print ra0,dec0
        ra0, dec0 = wcs.pix2wcs(ra0, dec0)
        ra0 *= DEG
        dec0 *= DEG
        #    print ModelClasses.Position.ra_hms_static(ra0)
        #    print ModelClasses.Position.dec_sdms_static(dec0)
        sx, sy = wcs.getHalfSizeDeg()
        sx *= DEG
        sy *= DEG
        nx, ny = input_hdu.data.shape[-1:-3:-1]
        pos = ModelClasses.Position(ra0, dec0)
        flux = ModelClasses.Flux(max_flux)
        shape = ModelClasses.FITSImage(sx, sy, 0, os.path.basename(filename), nx, ny,
                                       pad=float(str(self.wpad.text()) or "1"))
        img_src = SkyModel.Source(srcname, pos, flux, shape=shape)
        self.model.setSources(self.model.sources + [img_src])
        self.model.emitUpdate(SkyModel.SkyModel.UpdateAll, origin=self)
        busy.reset_cursor()
        return QDialog.accept(self)


def add_brick(mainwin, model):
    dialog = getattr(mainwin, '_add_brick_dialog', None)
    if not dialog:
        dialog = mainwin._add_brick_dialog = AddBrickDialog(mainwin)
    dialog.setModel(model)
    # show dialog
    return dialog.exec_()


from TigGUI.Tools import registerTool

registerTool("Add FITS brick to model...", add_brick)



# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os.path

FITS_ExtensionList = [".fts", ".FTS", ".fits", ".FITS", ".fit", ".FIT"]


def isFITS(filename):
    return os.path.splitext(filename)[1] in FITS_ExtensionList



# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import traceback

import numpy
from PyQt5.Qt import QHBoxLayout, QFileDialog, QComboBox, QLabel, QLineEdit, QDialog, QToolButton, \
    Qt, QApplication, QColor, QPixmap, QPainter, QFrame, QMenu, QPen, QKeySequence, QCheckBox
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QDockWidget, QSizePolicy, QWidget, QPushButton, QStyle
from PyQt5.Qwt import QwtText, QwtPlotCurve, QwtPlotMarker, QwtScaleMap, QwtPlotItem
from PyQt5.QtCore import pyqtSignal, QPoint, QPointF, QSize

import TigGUI.kitties.utils
from TigGUI.Images.SkyImage import FITSImagePlotItem
from TigGUI.Plot.SkyModelPlot import LiveImageZoom
from TigGUI.kitties.utils import PersistentCurrier
from TigGUI.kitties.widgets import BusyIndicator


QStringList = list

_verbosity = TigGUI.kitties.utils.verbosity(name="imagectl")
dprint = _verbosity.dprint
dprintf = _verbosity.dprintf

from TigGUI.init import pixmaps
from TigGUI.Widgets import FloatValidator, TDockWidget
from TigGUI.Images.RenderControl import RenderControl
from TigGUI.Images.ControlDialog import ImageControlDialog


class ImageController(QFrame):
    """An ImageController is a widget for controlling the display of one image.
    It can emit the following signals from the image:
    raise                     raise button was clicked
    center                  center-on-image option was selected
    unload                  unload option was selected
    slice                     image slice has changed, need to redraw (emitted by SkyImage automatically)
    repaint                 image display range or colormap has changed, need to redraw (emitted by SkyImage automatically)
    """

    # image signals
    imageSignalRepaint = pyqtSignal()
    imageSignalSlice = pyqtSignal(tuple)
    imageSignalRaise = pyqtSignal([FITSImagePlotItem])
    imageSignalUnload = pyqtSignal(object)
    imageSignalCenter = pyqtSignal(object)

    def __init__(self, image, parent, imgman, name=None, save=False):
        QFrame.__init__(self, parent)
        self.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
        # init state
        self._border_pen = None
        self._image_label_text = None
        self._subset = None
        self.image = image
        self._imgman = imgman
        self._currier = PersistentCurrier()
        self._control_dialog = None
        # create widgets
        self._lo = lo = QHBoxLayout(self)
        lo.setContentsMargins(0, 0, 0, 0)
        lo.setSpacing(2)
        # raise button
        self._wraise = QToolButton(self)
        lo.addWidget(self._wraise)
        self._wraise.setIcon(pixmaps.raise_up.icon())
        self._wraise.setAutoRaise(True)
        self._can_raise = False
        self._wraise.clicked.connect(self._raiseButtonPressed)
        self._wraise.setToolTip("""<P>Click here to raise this image above other images. Hold the button down briefly to
      show a menu of image operations.</P>""")
        # center label
        self._wcenter = QLabel(self)
        self._wcenter.setPixmap(pixmaps.center_image.pm())
        self._wcenter.setToolTip(
            "<P>The plot is currently centered on (the reference pixel %d,%d) of this image.</P>" % self.image.referencePixel())
        lo.addWidget(self._wcenter)
        # name/filename label
        self.name = image.name
        self._wlabel = QLabel(self.name, self)
        self._number = 0
        self.setName(self.name)
        self._wlabel.setToolTip("%s %s" % (image.filename, "\u00D7".join(map(str, image.data().shape))))
        lo.addWidget(self._wlabel, 1)
        # if 'save' is specified, create a "save" button
        if save:
            self._wsave = QToolButton(self)
            lo.addWidget(self._wsave)
            self._wsave.setText("save")
            self._wsave.setAutoRaise(True)
            self._save_dir = save if isinstance(save, str) else "."
            self._wsave.clicked.connect(self._saveImage)
            self._wsave.setToolTip("""<P>Click here to write this image to a FITS file.</P>""")
        # render control
        self.image.connectRepaint(self.imageSignalRepaint)
        self.image.connectSlice(self.imageSignalSlice)
        self.image.connectRaise(self.imageSignalRaise)
        self.image.connectUnload(self.imageSignalUnload)
        self.image.connectCenter(self.imageSignalCenter)
        dprint(2, "creating RenderControl")
        self._rc = RenderControl(image, self)
        dprint(2, "done")
        # selectors for extra axes
        self._wslicers = []
        curslice = self._rc.currentSlice()  # this may be loaded from config, so not necessarily 0
        for iextra, axisname, labels in self._rc.slicedAxes():
            if axisname.upper() not in ["STOKES", "COMPLEX"]:
                lbl = QLabel("%s:" % axisname, self)
                lo.addWidget(lbl)
            else:
                lbl = None
            slicer = QComboBox(self)
            self._wslicers.append(slicer)
            lo.addWidget(slicer)
            slicer.addItems(labels)
            slicer.setToolTip("""<P>Selects current slice along the %s axis.</P>""" % axisname)
            slicer.setCurrentIndex(curslice[iextra])
            slicer.activated[int].connect(self._currier.curry(self._rc.changeSlice, iextra))
        # min/max display ranges
        lo.addSpacing(5)
        self._wrangelbl = QLabel(self)
        lo.addWidget(self._wrangelbl)
        self._minmaxvalidator = FloatValidator(self)
        self._wmin = QLineEdit(self)
        self._wmax = QLineEdit(self)
        width = self._wmin.fontMetrics().width("1.234567e-05")
        for w in self._wmin, self._wmax:
            lo.addWidget(w, 0)
            w.setValidator(self._minmaxvalidator)
            w.setMaximumWidth(width)
            w.setMinimumWidth(width)
            w.editingFinished.connect(self._changeDisplayRange)
        # full-range button
        self._wfullrange = QToolButton(self)
        lo.addWidget(self._wfullrange, 0)
        self._wfullrange.setIcon(pixmaps.zoom_range.icon())
        self._wfullrange.setAutoRaise(True)
        self._wfullrange.clicked.connect(self.renderControl().resetSubsetDisplayRange)
        rangemenu = QMenu(self)
        rangemenu.addAction(pixmaps.full_range.icon(), "Full subset", self.renderControl().resetSubsetDisplayRange)
        for percent in (99.99, 99.9, 99.5, 99, 98, 95):
            rangemenu.addAction("%g%%" % percent, self._currier.curry(self._changeDisplayRangeToPercent, percent))
        self._wfullrange.setPopupMode(QToolButton.DelayedPopup)
        self._wfullrange.setMenu(rangemenu)
        # update widgets from current display range
        self._updateDisplayRange(*self._rc.displayRange())
        # lock button
        self._wlock = QToolButton(self)
        self._wlock.setIcon(pixmaps.unlocked.icon())
        self._wlock.setAutoRaise(True)
        self._wlock.setToolTip("""<P>Click to lock or unlock the intensity range. When the intensity range is locked across multiple images, any changes in the intensity
          range of one are propagated to the others. Hold the button down briefly for additional options.</P>""")
        lo.addWidget(self._wlock)
        self._wlock.clicked.connect(self._toggleDisplayRangeLock)
        self.renderControl().displayRangeLocked.connect(self._setDisplayRangeLock)
        self.renderControl().dataSubsetChanged.connect(self._dataSubsetChanged)
        lockmenu = QMenu(self)
        lockmenu.addAction(pixmaps.locked.icon(), "Lock all to this",
                           self._currier.curry(imgman.lockAllDisplayRanges, self.renderControl()))
        lockmenu.addAction(pixmaps.unlocked.icon(), "Unlock all", imgman.unlockAllDisplayRanges)
        self._wlock.setPopupMode(QToolButton.DelayedPopup)
        self._wlock.setMenu(lockmenu)
        self._setDisplayRangeLock(self.renderControl().isDisplayRangeLocked())
        # dialog button
        self._wshowdialog = QToolButton(self)
        lo.addWidget(self._wshowdialog)
        self._wshowdialog.setIcon(pixmaps.colours.icon())
        self._wshowdialog.setAutoRaise(True)
        self._wshowdialog.setToolTip("""<P>Click for colourmap and intensity policy options.</P>""")
        self._wshowdialog.clicked.connect(self.showRenderControls)
        tooltip = """<P>You can change the currently displayed intensity range by entering low and high limits here.</P>
            <TABLE>
            <TR><TD><NOBR>Image min:</NOBR></TD><TD>%g</TD><TD>max:</TD><TD>%g</TD></TR>
            </TABLE>""" % self.image.imageMinMax()
        for w in self._wmin, self._wmax, self._wrangelbl:
            w.setToolTip(tooltip)
        # create image operations menu
        self._menu = QMenu(self.name, self)
        self._qa_raise = self._menu.addAction(pixmaps.raise_up.icon(), "Raise image",
                                              self._currier.curry(self.image.signalRaise.emit, None))
        self._qa_center = self._menu.addAction(pixmaps.center_image.icon(), "Center plot on image",
                                               self._currier.curry(self.image.signalCenter.emit, True))
        self._qa_show_rc = self._menu.addAction(pixmaps.colours.icon(), "Colours && Intensities...",
                                                self.showRenderControls)
        if save:
            self._qa_save = self._menu.addAction("Save image...", self._saveImage)
        self._menu.addAction("Export image to PNG file...", self._exportImageToPNG)
        self._export_png_dialog = None
        self._menu.addAction("Unload image", self._currier.curry(self.image.signalUnload.emit, None))
        self._wraise.setMenu(self._menu)
        self._wraise.setPopupMode(QToolButton.DelayedPopup)
        # connect updates from renderControl and image
        self.image.signalSlice.connect(self._updateImageSlice)
        self._rc.displayRangeChanged.connect(self._updateDisplayRange)
        # default plot depth of image markers
        self._z_markers = None
        # and the markers themselves
        self._image_border = QwtPlotCurve()
        self._image_border.setRenderHint(QwtPlotItem.RenderAntialiased)
        self._image_label = QwtPlotMarker()
        self._image_label.setRenderHint(QwtPlotItem.RenderAntialiased)
        # subset markers
        self._subset_pen = QPen(QColor("Light Blue"))
        self._subset_border = QwtPlotCurve()
        self._subset_border.setRenderHint(QwtPlotItem.RenderAntialiased)
        self._subset_border.setPen(self._subset_pen)
        self._subset_border.setVisible(False)
        self._subset_label = QwtPlotMarker()
        self._subset_label.setRenderHint(QwtPlotItem.RenderAntialiased)
        text = QwtText("subset")
        text.setColor(self._subset_pen.color())
        self._subset_label.setLabel(text)
        self._subset_label.setLabelAlignment(Qt.AlignRight | Qt.AlignBottom)
        self._subset_label.setVisible(False)
        self._setting_lmrect = False
        self._all_markers = [self._image_border, self._image_label, self._subset_border, self._subset_label]
        self._exportMaxRes = False
        self._dockable_colour_ctrl = None

    def close(self):
        if self._control_dialog:
            self._control_dialog.close()
            self._control_dialog = None

    def __del__(self):
        self.close()

    def __eq__(self, other):
        return self is other

    def renderControl(self):
        return self._rc

    def getMenu(self):
        return self._menu

    def getFilename(self):
        return self.image.filename

    def setName(self, name):
        self.name = name
        self._wlabel.setText("%s: %s" % (chr(ord('a') + self._number), self.name))

    def setNumber(self, num):
        self._number = num
        self._menu.menuAction().setText("%s: %s" % (chr(ord('a') + self._number), self.name))
        self._qa_raise.setShortcut(QKeySequence("Alt+" + chr(ord('A') + num)))
        self.setName(self.name)

    def getNumber(self):
        return self._number

    def setPlotProjection(self, proj):
        self.image.setPlotProjection(proj)
        sameproj = proj == self.image.projection
        self._wcenter.setVisible(sameproj)
        self._qa_center.setVisible(not sameproj)
        if self._image_border:
            (l0, l1), (m0, m1) = self.image.getExtents()
            path = numpy.array([l0, l0, l1, l1, l0]), numpy.array([m0, m1, m1, m0, m0])
            self._image_border.setSamples(*path)
            if self._image_label:
                self._image_label.setValue(path[0][2], path[1][2])

    def addPlotBorder(self, border_pen, label, label_color=None, bg_brush=None):
        # make plot items for image frame
        # make curve for image borders
        (l0, l1), (m0, m1) = self.image.getExtents()
        self._border_pen = QPen(border_pen)
        self._image_border.show()
        self._image_border.setSamples([l0, l0, l1, l1, l0], [m0, m1, m1, m0, m0])
        self._image_border.setPen(self._border_pen)
        self._image_border.setZ(self.image.z() + 1 if self._z_markers is None else self._z_markers)
        if label:
            self._image_label.show()
            self._image_label_text = text = QwtText(" %s " % label)
            text.setColor(label_color)
            text.setBackgroundBrush(bg_brush)
            self._image_label.setValue(l1, m1)
            self._image_label.setLabel(text)
            self._image_label.setLabelAlignment(Qt.AlignRight | Qt.AlignVCenter)
            self._image_label.setZ(self.image.z() + 2 if self._z_markers is None else self._z_markers)

    def setPlotBorderStyle(self, border_color=None, label_color=None):
        if border_color:
            self._border_pen.setColor(border_color)
            self._image_border.setPen(self._border_pen)
        if label_color:
            self._image_label_text.setColor(label_color)
            self._image_label.setLabel(self._image_label_text)

    def showPlotBorder(self, show=True):
        self._image_border.setVisible(show)
        self._image_label.setVisible(show)

    def attachToPlot(self, plot, z_markers=None):
        for item in [self.image] + self._all_markers:
            if item.plot() != plot:
                item.attach(plot)

    def setImageVisible(self, visible):
        self.image.setVisible(visible)

    def showRenderControls(self):
        if not self._control_dialog:
            dprint(1, "creating control dialog")
            self._control_dialog = ImageControlDialog(self, self._rc, self._imgman)
            # line below allows window to be resized by the user
            self._control_dialog.setSizeGripEnabled(True)
            # get and set sizing
            self._control_dialog.setMinimumWidth(396)
            # create size policy for control dialog
            colour_ctrl_policy = QSizePolicy()
            colour_ctrl_policy.setHorizontalPolicy(QSizePolicy.Minimum)
            self._control_dialog.setSizePolicy(colour_ctrl_policy)
            # setup dockable colour control dialog
            self._dockable_colour_ctrl = TDockWidget(title=f"{self._rc.image.name}", parent=self.parent().mainwin,
                                                     bind_widget=self._control_dialog,
                                                     close_slot=self.colourctrl_dockwidget_closed,
                                                     toggle_slot=self.colourctrl_dockwidget_toggled)
            self.addDockWidgetToTab()
            dprint(1, "done")
        # set dockable widget visibility in sync with control dialog
        if not self._control_dialog.isVisible():
            dprint(1, "showing control dialog")
            self._control_dialog.show()
            if self._dockable_colour_ctrl is not None:
                self._dockable_colour_ctrl.setVisible(True)
                self.addDockWidgetToTab()
                self._dockable_colour_ctrl.show()
                self._dockable_colour_ctrl.raise_()
                if not self.get_docked_widget_size(self._dockable_colour_ctrl):
                    geo = self.parent().mainwin.geometry()
                    geo.setWidth(self.parent().mainwin.width() + self._dockable_colour_ctrl.width())
                    self.parent().mainwin.setGeometry(geo)
        else:
            self._control_dialog.hide()
            self._dockable_colour_ctrl.setVisible(False)
            if not self.get_docked_widget_size(self._dockable_colour_ctrl):
                geo = self.parent().mainwin.geometry()
                geo.setWidth(self.parent().mainwin.width() - self._dockable_colour_ctrl.width())
                self.parent().mainwin.setGeometry(geo)

    def addDockWidgetToTab(self):
        # Add dockable widget to main window.
        # This needs to itterate through the widgets to find DockWidgets already in the right side area,
        # then tabifydockwidget when adding, or add to the right area if empty
        widget_list = self.parent().mainwin.findChildren(QDockWidget)
        for widget in widget_list:
            if self.parent().mainwin.dockWidgetArea(widget) == 2:  # if in right dock area
                if widget.isVisible() and not widget.isFloating():  # if widget active and not a window
                    if self._dockable_colour_ctrl is not widget:  # check not itself
                        # add dock widget in tab on top of current widget in right area
                        self.parent().mainwin.tabifyDockWidget(widget, self._dockable_colour_ctrl)
                        self.parent().mainwin.resizeDocks([widget], [widget.bind_widget.width()], Qt.Horizontal)
            elif self.parent().mainwin.dockWidgetArea(
                    widget) == 0:  # if not in any dock area assume we have new dock widget
                # no previous widget in this area then add
                self.parent().mainwin.addDockWidget(Qt.RightDockWidgetArea, self._dockable_colour_ctrl)
                self.parent().mainwin.resizeDocks([widget], [widget.bind_widget.width()], Qt.Horizontal)

    def removeDockWidget(self):
        # remove image control dock widget
        self.parent().mainwin.removeDockWidget(self._dockable_colour_ctrl)
        # get widgets to resize
        widget_list = self.parent().mainwin.findChildren(QDockWidget)
        size_list = []
        result = []
        for widget in widget_list:
            if not isinstance(widget.bind_widget, ImageControlDialog):
                size_list.append(widget.bind_widget.width())
                result.append(widget)
                dprint(2, f"{widget} width {widget.width()}")
                dprint(2, f"{widget} bind_widget width {widget.bind_widget.width()}")
                if isinstance(widget.bind_widget, LiveImageZoom):
                    widget.bind_widget.setMinimumWidth(widget.width())
        widget_list = result
        # resize dock areas
        self.parent().mainwin.resizeDocks(widget_list, size_list, Qt.Horizontal)

    def colourctrl_dockwidget_closed(self):
        self._dockable_colour_ctrl.setVisible(False)
        if self.parent().mainwin.windowState() != Qt.WindowMaximized:
            if not self.get_docked_widget_size(self._dockable_colour_ctrl):
                if not self._dockable_colour_ctrl.isFloating():
                    geo = self.parent().mainwin.geometry()
                    geo.setWidth(self.parent().mainwin.width() - self._dockable_colour_ctrl.width())
                    self.parent().mainwin.setGeometry(geo)

    def colourctrl_dockwidget_toggled(self):
        if not self._dockable_colour_ctrl.isVisible():
            return
        if self._dockable_colour_ctrl.isWindow():
            self._dockable_colour_ctrl.setFloating(False)
            if self.parent().mainwin.windowState() != Qt.WindowMaximized:
                if not self.get_docked_widget_size(self._dockable_colour_ctrl):
                    geo = self.parent().mainwin.geometry()
                    geo.setWidth(self.parent().mainwin.width() + self._dockable_colour_ctrl.width())
                    self.parent().mainwin.setGeometry(geo)
        else:
            self._dockable_colour_ctrl.setFloating(True)
            if self.parent().mainwin.windowState() != Qt.WindowMaximized:
                if not self.get_docked_widget_size(self._dockable_colour_ctrl):
                    geo = self.parent().mainwin.geometry()
                    geo.setWidth(self.parent().mainwin.width() - self._dockable_colour_ctrl.width())
                    self.parent().mainwin.setGeometry(geo)

    def get_docked_widget_size(self, _dockable):
        widget_list = self.parent().mainwin.findChildren(QDockWidget)
        size_list = []
        if _dockable:
            for widget in widget_list:
                if isinstance(widget.bind_widget, ImageControlDialog):
                    if widget is not _dockable:
                        if not widget.isWindow() and not widget.isFloating() and widget.isVisible():
                            size_list.append(widget.bind_widget.width())
        if size_list:
            return max(size_list)
        else:
            return size_list

    def _changeDisplayRangeToPercent(self, percent):
        if not self._control_dialog:
            self._control_dialog = ImageControlDialog(self, self._rc, self._imgman)
        self._control_dialog._changeDisplayRangeToPercent(percent)

    def _updateDisplayRange(self, dmin, dmax):
        """Updates display range widgets."""
        self._wmin.setText("%.4g" % dmin)
        self._wmax.setText("%.4g" % dmax)
        self._updateFullRangeIcon()

    def _changeDisplayRange(self):
        """Gets display range from widgets and updates the image with it."""
        try:
            newrange = float(str(self._wmin.text())), float(str(self._wmax.text()))
        except ValueError:
            return
        self._rc.setDisplayRange(*newrange)

    def _dataSubsetChanged(self, subset, minmax, desc, subset_type):
        """Called when the data subset changes (or is reset)"""
        # hide the subset indicator -- unless we're invoked while we're actually setting the subset itself
        if not self._setting_lmrect:
            self._subset = None
            self._subset_border.setVisible(False)
            self._subset_label.setVisible(False)

    def setLMRectSubset(self, rect):
        self._subset = rect
        l0, m0, l1, m1 = rect.getCoords()
        self._subset_border.setSamples([l0, l0, l1, l1, l0], [m0, m1, m1, m0, m0])
        self._subset_border.setVisible(True)
        self._subset_label.setValue(max(l0, l1), max(m0, m1))
        self._subset_label.setVisible(True)
        self._setting_lmrect = True
        self.renderControl().setLMRectSubset(rect)
        self._setting_lmrect = False

    def currentSlice(self):
        return self._rc.currentSlice()

    def _updateImageSlice(self, _slice):
        dprint(2, _slice)
        for i, (iextra, name, labels) in enumerate(self._rc.slicedAxes()):
            slicer = self._wslicers[i]
            if slicer.currentIndex() != _slice[iextra]:
                dprint(3, "setting widget", i, "to", _slice[iextra])
                slicer.setCurrentIndex(_slice[iextra])

    def setMarkersZ(self, z):
        self._z_markers = z
        for i, elem in enumerate(self._all_markers):
            elem.setZ(z + i)

    def setZ(self, z, top=False, depthlabel=None, can_raise=True):
        self.image.setZ(z)
        if self._z_markers is None:
            for i, elem in enumerate(self._all_markers):
                elem.setZ(z + i + i)
        # set the depth label, if any
        label = "%s: %s" % (chr(ord('a') + self._number), self.name)
        # label = "%s %s"%(depthlabel,self.name) if depthlabel else self.name
        if top:
            label = "%s: <B>%s</B>" % (chr(ord('a') + self._number), self.name)
        self._wlabel.setText(label)
        # set hotkey
        self._qa_show_rc.setShortcut(Qt.Key_F9 if top else QKeySequence())
        # set raise control
        self._can_raise = can_raise
        self._qa_raise.setVisible(can_raise)
        self._wlock.setVisible(can_raise)
        if can_raise:
            self._wraise.setToolTip(
                "<P>Click here to raise this image to the top. Click on the down-arrow to access the image menu.</P>")
        else:
            self._wraise.setToolTip("<P>Click to access the image menu.</P>")

    def _raiseButtonPressed(self):
        if self._can_raise:
            self.image.signalRaise.emit(self.image)
        else:
            self._wraise.showMenu()

    def _saveImage(self):
        filename = QFileDialog.getSaveFileName(self, "Save FITS file", self._save_dir,
                                               "FITS files(*.fits *.FITS *fts *FTS)", options=QFileDialog.DontUseNativeDialog)
        filename = str(filename[0])
        if not filename:
            return
        busy = BusyIndicator()
        self._imgman.signalShowMessage.emit("""Writing FITS image %s""" % filename, 3000)
        QApplication.flush()
        try:
            self.image.save(filename)
        except Exception as exc:
            busy.reset_cursor()
            traceback.print_exc()
            self._imgman.signalShowErrorMessage.emit("""Error writing FITS image %s: %s""" % (filename, str(sys.exc_info()[1])))
            return None
        self.renderControl().startSavingConfig(filename)
        self.setName(self.image.name)
        self._qa_save.setVisible(False)
        self._wsave.hide()
        busy.reset_cursor()

    def _exportImageResolution(self):
        sender = self.sender()
        if isinstance(sender, QCheckBox):
            if sender.isChecked():
                self._exportMaxRes = True
            else:
                self._exportMaxRes = False

    def _exportImageToPNG(self, filename=None):
        if not filename:
            if not self._export_png_dialog:
                dialog = self._export_png_dialog = QFileDialog(self, "Export image to PNG", ".", "*.png")
                dialog.setDefaultSuffix("png")
                dialog.setFileMode(QFileDialog.AnyFile)
                dialog.setAcceptMode(QFileDialog.AcceptSave)
                dialog.setModal(True)
                dialog.filesSelected['QStringList'].connect(self._exportImageToPNG)
                # attempt to add limit 4K option - not available on Ubuntu Unity
                layout = dialog.layout()
                if layout is not None:
                    checkbox = QCheckBox("Limit to 4K image")
                    checkbox.setChecked(False)
                    checkbox.setToolTip("Limits the image output to 4K")
                    checkbox.toggled.connect(self._exportImageResolution)
                    layout.addWidget(checkbox)
                    dialog.setLayout(layout)
            return self._export_png_dialog.exec_() == QDialog.Accepted
        busy = BusyIndicator()
        if isinstance(filename, QStringList):
            filename = filename[0]
        filename = str(filename)
        # get image dimensions
        nx, ny = self.image.imageDims()
        # export either max resolution possible or default to 4K. If image is small then no scaling occurs.
        if not self._exportMaxRes:
            # get free memory. Note: Linux only!
            import os
            total_memory, used_memory, free_memory = map(int, os.popen('free -t -m').readlines()[-1].split()[1:])
            # use 90% of free memory available
            free_memory = free_memory * 0.9
            # use an approximation to find the max image size that can be generated
            if nx >= ny and nx > free_memory:
                scale_factor = round(free_memory / nx, 1)
            elif ny > nx and ny > free_memory:
                scale_factor = round(free_memory / ny, 1)
            else:
                scale_factor = 1
        else:
            # default to 4K
            if nx > 4000:
                scale_factor = 4000 / nx
            elif ny > nx and ny > 4000:
                scale_factor = 4000 / ny
            else:
                scale_factor = 1

        # make QPixmap
        nx = nx * scale_factor
        ny = ny * scale_factor
        (l0, l1), (m0, m1) = self.image.getExtents()
        pixmap = QPixmap(nx, ny)
        painter = QPainter(pixmap)
        # use QwtPlot implementation of draw canvas, since we want to avoid caching
        xmap = QwtScaleMap()
        xmap.setPaintInterval(0, nx)
        xmap.setScaleInterval(l1, l0)
        ymap = QwtScaleMap()
        ymap.setPaintInterval(ny, 0)
        ymap.setScaleInterval(m0, m1)
        # call painter with clear cache option for consistent file size output.
        self.image.draw(painter, xmap, ymap, pixmap.rect(), use_cache=False)
        painter.end()
        # save to file
        try:
            pixmap.save(filename, "PNG")
            # clean up export items
            pixmap.detach()
            del xmap
            del ymap
            del pixmap
            del painter
        except Exception as exc:
            self._imgman.signalShowErrorMessage[str, int].emit("Error writing %s: %s" % (filename, str(exc)), 3000)
            busy.reset_cursor()
        else:
            busy.reset_cursor()
            self._imgman.signalShowMessage[str, int].emit("Exported image to file %s" % filename, 3000)

    def _toggleDisplayRangeLock(self):
        self.renderControl().lockDisplayRange(not self.renderControl().isDisplayRangeLocked())

    def _setDisplayRangeLock(self, locked):
        self._wlock.setIcon(pixmaps.locked.icon() if locked else pixmaps.unlocked.icon())

    def _updateFullRangeIcon(self):
        if self._rc.isSubsetDisplayRange():
            self._wfullrange.setIcon(pixmaps.zoom_range.icon())
            self._wfullrange.setToolTip(
                """<P>The current intensity range is the full range. Hold this button down briefly for additional options.</P>""")
        else:
            self._wfullrange.setIcon(pixmaps.full_range.icon())
            self._wfullrange.setToolTip(
                """<P>Click to reset to a full intensity range. Hold the button down briefly for additional options.</P>""")



# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""Checks that Tigger-LSM, PyQt5 and PyQt-Qwt are available for Tigger to operate."""
missing = []  # list of missing dependencies


def close_btn(btn):
    """QMessageBox close button handler to close GUI and print error message to terminal."""
    msg_box.close()
    sys.exit(error_msg)


try:
    try:
        from Tigger.Models import ModelClasses  # check Tigger-LSM is available
    except ImportError:
        missing.append("Tigger-LSM")
        pass

    try:
        from PyQt5.Qt import Qt  # check PyQt5 is available
    except ImportError:
        missing.append("PyQt5")
        pass

    try:
        from PyQt5.QtOpenGL import QGLWidget  # check PyQt5 Qt OpenGL is available
    except ImportError:
        missing.append("PyQt5.QtOpenGL")
        pass

    try:
        from PyQt5.QtSvg import QSvgWidget  # check PyQt5 Qt SVG is available
    except ImportError:
        missing.append("PyQt5.QtSvg")
        pass

    try:
        from PyQt5.Qwt import QwtPlotZoomer  # check PyQt-Qwt is available

        test_qwt = callable(
            getattr(QwtPlotZoomer, 'setZoomStack', False))  # check correct version of PyQt-Qwt is installed
    except ImportError:
        missing.append("PyQt5.Qwt")
        test_qwt = False
        pass

except ImportError:
    deps_available = False
    pass
else:
    if test_qwt:
        deps_available = True
    else:
        deps_available = False
        missing.append("PyQt-Qwt>=2.0.0")

if not deps_available:
    import sys

    error_msg = f"Error: Dependencies have not been met {missing}, please check your installation. \n" \
                "See https://github.com/ratt-ru/tigger for further information."

    # load GUI error message if possible
    if 'PyQt5' not in missing:
        from PyQt5.Qt import QMessageBox, QApplication, QPalette, QColor

        # set dark theme palatte
        palette = QPalette()
        palette.setColor(QPalette.Window, QColor(53, 53, 53))
        palette.setColor(QPalette.WindowText, Qt.white)
        palette.setColor(QPalette.Light, QColor(68, 68, 68))
        palette.setColor(QPalette.Base, QColor(25, 25, 25))
        palette.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
        palette.setColor(QPalette.ToolTipBase, Qt.white)
        palette.setColor(QPalette.ToolTipText, Qt.black)
        palette.setColor(QPalette.Text, Qt.white)
        palette.setColor(QPalette.Button, QColor(53, 53, 53))
        palette.setColor(QPalette.ButtonText, Qt.white)
        palette.setColor(QPalette.BrightText, Qt.red)
        palette.setColor(QPalette.Link, QColor(42, 130, 218))
        palette.setColor(QPalette.Highlight, QColor(42, 130, 218, 192))
        palette.setColor(QPalette.HighlightedText, Qt.white)

        # create default QApp and set dark theme palette
        app = QApplication(sys.argv)
        app.setPalette(palette)

        # create GUI dialog window to display error message
        msg_box = QMessageBox()
        msg_box.setWindowFlag(Qt.CustomizeWindowHint, True)
        msg_box.setWindowFlag(Qt.WindowCloseButtonHint, False)
        msg_box.setTextFormat(Qt.RichText)
        msg_box.setIcon(QMessageBox.Warning)
        msg_box.setWindowTitle("Tigger")
        msg_box.setText(f"<b>Error</b><br/><br/>Dependencies have not been met: <br/><br/>{missing}<br/><br/>"
                        f"Please check your installation. "
                        f"<br/><br/>See <a href='https://github.com/ratt-ru/tigger'>https://github.com/ratt-ru/tigger</a>"
                        f" for further information.")
        msg_box.setStandardButtons(QMessageBox.Close)
        msg_box.buttonClicked.connect(close_btn)
        msg_box.exec_()
    else:
        sys.exit(error_msg)
